import curses
import todoism.utils as tutils
import todoism.task as tsk
import todoism.print as tprint


def main(_):
    global task_highlighting_color
    stdscr = curses.initscr()
    stdscr.keypad(True) # enable e.g arrow keys
    stdscr.scrollok(True)
    curses.curs_set(1)
    stdscr.clear()
    stdscr.refresh()
    # Initialize color pairs
    curses.start_color()
    # progress colors
    curses.init_pair(2, curses.COLOR_GREEN, curses.COLOR_BLACK)
    curses.init_pair(3, curses.COLOR_YELLOW, curses.COLOR_BLACK)
    curses.init_pair(4, curses.COLOR_RED, curses.COLOR_BLACK)
    curses.init_pair(5, curses.COLOR_WHITE, curses.COLOR_BLACK) # regular color pair
    # Set up the screen
    curses.curs_set(0)
    stdscr.clear()
    # Assuming color pair 0 represents the default colors
    stdscr.bkgd(' ', curses.COLOR_BLACK | curses.A_NORMAL)

    # Define the initial todo list
    task_list = tsk.load_tasks()
    done_list = [] # a  part of task list
    purged_list = []

    current_id = 1
    current_row = 1
    tutils.reid(task_list) # reid in case something went wrong in last session
    task_cnt = len(task_list) # done + undone
    done_cnt = tsk.done_count(task_list)
    window_height = stdscr.getmaxyx()[0]    
    # print window of task id
    start = 1 if task_cnt > 0 else 0
    end = task_cnt if task_cnt < window_height - 1 else window_height - 1
    
        
    while True:
        task_cnt = len(task_list)
        done_cnt = tsk.done_count(task_list)

        stdscr.clear()
        # Selected task highlighting
        color_selected = tutils.get_color_selected()
        curses.init_pair(1, curses.COLOR_BLACK, color_selected)

        tprint.print_status_bar(stdscr, done_cnt, task_cnt)
        tprint.print_tasks(stdscr, task_list, current_id, start, end)
        stdscr.refresh()
        window_height = stdscr.getmaxyx()[0]
        
        # for restoring previous view if add interrupted
        old_start = start
        old_end = end
        # Wait for user input
        key = stdscr.getch()
        # Handle user input
        if key == ord('a'):
            curses.echo()
            curses.curs_set(1)
            # adjust start end for pre-print
            # taskoverflow if a new one is added: 
            if task_cnt >= window_height - 1:
                if end <= task_cnt:
                    start = task_cnt - (end - start - 1)
                    end = task_cnt
                else:
                    pass
            stdscr.erase()
            tprint.print_status_bar(stdscr, done_cnt, task_cnt)
            tprint.print_tasks(stdscr, task_list, current_id, start, end)
            stdscr.addstr(window_height - 1 if task_cnt >= window_height - 1 else task_cnt + 1, 4 if task_cnt < 9 else 3, f"{task_cnt + 1}.{' '}")
            stdscr.refresh()
            
            # Add a new task
            new_task_description = tutils.edit(stdscr, tsk.create_new_task(task_cnt + 1), tprint.add_mode)               
            if new_task_description != "":
                new_id = task_cnt + 1
                task_list = tsk.add_new_task(task_list, new_id, new_task_description)
                task_cnt = task_cnt + 1
                if task_cnt == 1:
                    start = 1
                current_id = new_id # new id
                if end == task_cnt:
                    current_row = current_row + 1
                else:
                    current_row = window_height - 1
                end = end + 1 # change end as well
            else:
                start = old_start
                end = old_end
            tprint.print_tasks(stdscr, task_list, current_id, start, end)
            stdscr.refresh()  
            curses.curs_set(0)
            curses.noecho()
        elif key == ord("d"):
            # mark the current task as 'done'
            if task_list:
                done_list.append(task_list[current_id - 1])
                task_list[current_id - 1]['status'] = not task_list[current_id - 1]['status']  
                done_cnt = done_cnt + 1 if task_list[current_id - 1]['status'] is True else done_cnt - 1
                tsk.save_tasks(task_list, tsk.tasks_file_path)
        elif key == ord('e'):
            curses.echo()
            curses.curs_set(1)
            if task_cnt > 0:
                stdscr.move(current_id if current_id <= window_height - 1 else window_height - 1, len(task_list[current_id - 1]['description']) + tutils.indent)
                task_list[current_id - 1]['description'] = tutils.edit(stdscr, task_list[current_id - 1], tprint.edit_mode)
                # delete the task if it was edited to empty
                if task_list[current_id - 1]['description'] == "":
                    del task_list[current_id - 1]
                    tutils.reid(task_list)
                    if current_id > 1:
                        current_id = current_id - 1
                tsk.save_tasks(task_list, tsk.tasks_file_path)
            curses.curs_set(0)
            curses.noecho()        
        elif key == ord('f'):
            task_list[current_id - 1]['flagged'] = not task_list[current_id - 1]['flagged'] 
        elif key == ord('h'):
            stdscr.addstr(len(done_list) + 1, 0, "Completed Tasks:")
            for i, task in enumerate(done_list):
                stdscr.addstr(len(done_list) + i + 2, 0, f"{i + 1}. {task}")
            while stdscr.getch() is not ord('h'):
                continue
        elif key == ord('q'):
            break
        elif key == ord(':'):
            curses.echo()
            curses.curs_set(1)
            stdscr.addstr(curses.LINES - 1, 0, ":")
            stdscr.refresh()
            command_line = stdscr.getstr().decode('utf-8')
            curses.curs_set(0)
            curses.noecho()
            task_list, done_list, current_id, task_highlighting_color = tutils.execute_command(stdscr, command_line, task_list, done_list, purged_list, current_id)
            command_line = ""  # Clear the command line after executing the command
        elif key == curses.KEY_UP and current_id > 1:
            # current is top most task (id != 1)
            if task_cnt >= window_height - 1 and start > 1 and current_row == 1:
                start = start - 1
                end = end - 1
            else:
                current_row = current_row - 1
            current_id -= 1
        elif key == curses.KEY_DOWN and current_id < task_cnt:
            current_id += 1
            if task_cnt > window_height - 1 and current_row == window_height - 1:
                start = start + 1
                end = end + 1
            else:
                current_row = current_row + 1
        elif key == curses.KEY_BACKSPACE or key == 127:
            k = stdscr.getch() 
            if k == curses.KEY_BACKSPACE or k == 127:
                # only perform deletion when task_cnt > 0
                if task_cnt > 0:
                    # update done_cnt                      
                    if task_list[current_id - 1]['status'] is True:
                        done_cnt = done_cnt - 1
                    # perform deletion
                    del task_list[current_id - 1]
                    # reid
                    tutils.reid(task_list)
                    # view change rules are similar to apple reminder
                    if tutils.is_view_fully_packed(start, end, window_height - 1):
                        if window_height - 1 == task_cnt:
                            if current_id == end:
                                current_row = current_row - 1
                            end = end - 1
                            if current_id > 1:
                                current_id = current_id - 1
                        elif task_cnt == end and task_cnt > window_height - 1:
                            start = start - 1
                            end = end - 1
                            current_id = current_id - 1
                    else:
                        end = end - 1
                        if current_id == task_cnt:
                            current_row = current_row - 1
                            current_id = current_id - 1
             
                    # update task_cnt
                    task_cnt = task_cnt - 1                        
                tsk.save_tasks(task_list, tsk.tasks_file_path)
                # current_id = current_id - 1 if current_id > 1 else 1    

if __name__ == "__main__":
    curses.wrapper(main)



import curses
from datetime import datetime


add_mode  = 0
edit_mode = 1

help_msg =  '''
                         ┌──────────────────────────────────────────────────┐
                         │                                                  │
                         │   short commands:                                │
                         │   a - create new task                            │
                         │   d - mark task as done                          │
                         │   e - edit task                                  │
                         │   f - mark task as flagged                       │
                         │   q - quit this help message/todoism             │
                         │                                                  │
                         │   vim-like long commands:                        │            
                         │   (:<command> [args])                            │
                         │   :help - show this help message                 │
                         │   :purge - purge all done tasks                  │
                         │   :sort f - sort flagged tasks to top            │
                         │   :sort d - sort done tasks to bottom            │
                         │   :autosort f on|off                             │
                         │   :autosort d on|off                             │
                         │   :setcolor blue|red|yellow|green                │
                         │    - change background color of current task     │
                         │                                                  │
                         │   other key bindings:                            │
                         │   double Backspace - delete task                 │
                         │   ESC - quit adding/editing task                 │
                         │   Enter - finish adding/editing task             │
                         │   Up/Down Arrow Keys - navigate through tasks    │
                         │                                                  │
                         └──────────────────────────────────────────────────┘
            '''

def print_help(stdscr):
    height, width = stdscr.getmaxyx()
    stdscr.addstr(0, 0, help_msg)
    stdscr.refresh()
    # (width // 2) + 50
    # (height // 2) - 4

# The core function to print task
def print_task(stdscr, task, y):
    max_y = stdscr.getmaxyx()[0] 
    # handle task overflow
    if y < max_y:
        stdscr.addstr(y, 0, f"{'✅' if task['status'] else '  '} {' ' if task['id'] < 10 else ''}{task['id']}. {task['description'] + (75 - len(task['description'])) * ' ' + task['date']} {'🚩' if task['flagged'] else ''}" )

def print_task_highlighted(stdscr, task, y):
    stdscr.attron(curses.color_pair(1))
    print_task(stdscr, task, y)
    stdscr.attroff(curses.color_pair(1))        

def print_task_mode(stdscr, task, y, mode):
    if mode == edit_mode:
        print_task_highlighted(stdscr, task, y)
    else:
        print_task(stdscr, task, y)  
        
def print_tasks(stdscr, tasks, current_id, start, end):
    if start > 0:
        for i, task in enumerate(tasks[start - 1:end + 1]):
            if i + start == current_id: # handle task overflow: +start
                print_task_highlighted(stdscr, task, i + 1) # +1 due to status bar
            else:
                print_task(stdscr, task, i + 1)

def print_status_bar(stdscr, done_cnt, task_cnt):
    percentage_num = int((done_cnt / task_cnt) * 100) if task_cnt > 0 else 0
    status_bar = {
        'tasks': f'{' '*35}Progress: {done_cnt}/{task_cnt} {percentage_num if task_cnt > 0 else 0}%',
        'date': datetime.now().strftime("%Y-%m-%d %H:%M") 
    }
    
    color_pair = 0
    if percentage_num >= 67:
        color_pair = 2
    elif percentage_num >= 33:
        color_pair = 3
    else:
        if task_cnt == 0:
            color_pair = 5
        else:
            color_pair = 4
            
    stdscr.attron(curses.color_pair(color_pair))    
    stdscr.addstr(0, 0, f"{status_bar['tasks']}")
    stdscr.attroff(curses.color_pair(color_pair))
    stdscr.addstr(f" | {status_bar['date']}")
    stdscr.refresh()



import copy
import json
import curses
from datetime import datetime
import todoism.task as tsk
import todoism.print as tprint


indent = 7
description_length = 75
task_highlighting_color = curses.COLOR_BLUE

color_set = {
    "blue": curses.COLOR_BLUE,
    "red": curses.COLOR_RED,
    "yellow": curses.COLOR_YELLOW,
    "green": curses.COLOR_GREEN
}


def get_arg(argv):
    if len(argv) > 1:
        return argv[1]
    else:
        return ""


def reid(tasks):
    for i, t in enumerate(tasks):
        t['id'] = i + 1


def purge(tasks, purged_list):
    """
    purge completed tasks
    """
    remained = []
    for t in tasks:
        if t['status'] is False:
            remained.append(t)
        else:
            purged_list.append(t)
    reid(remained)
    tsk.save_tasks(purged_list, tsk.purged_file_path)
    return remained, []


def execute_command(stdscr, command, task_list, done_list, purged_list, current_id):
    global task_highlighting_color
    if command.startswith("add "):
        new_task = command[4:]
        if new_task:
            task_list.append({'description': new_task, 'date': datetime.now().strftime(
                "%Y-%m-%d %H:%M:%S"), 'status': False})
    elif command.startswith("done "):
        index_to_delete = int(command[7:]) - 1
        if 0 <= index_to_delete < len(task_list):
            done_list.append(copy.copy(task_list[index_to_delete]))
            task_list[index_to_delete]['status'] = not task_list[index_to_delete]['status']
    elif command == "purge":
        original_cnt = len(task_list)
        task_list, done_list = purge(task_list, purged_list)
        tsk.save_tasks(task_list, tsk.tasks_file_path)
        # change current id if some tasks were purged
        if len(task_list) < original_cnt:
            current_id = 1
    elif command.startswith("sort "):
        category = command[5:]
        if category == "f":
            flagged_tasks = []
            not_flagged = []
            for t in task_list:
                if t['flagged'] is True:
                    flagged_tasks.append(t)
                else:
                    not_flagged.append(t)
            task_list = flagged_tasks + not_flagged
            reid(task_list)
        elif category == 'd':
            done_tasks = []
            not_done = []
            for t in task_list:
                if t['status'] == True:
                    done_tasks.append(t)
                else:
                    not_done.append(t)
            task_list = not_done + done_tasks
            reid(task_list)
    elif command == "group":
        pass
    elif command.startswith("setcolor "):
        set_color_selected(command[9:])
    elif command == "help":
        tprint.print_help(stdscr)
        key = stdscr.getch()
        if key == ord('q'):
            stdscr.clear()
    # have to return task_highlighting_color, otherwise not accessible in main --> weird
    return task_list, done_list, current_id, task_highlighting_color


def edit(stdscr, task, mode):
    """
    A editing wrapper implemented using getch(). It delivers 
    more comprehensive functionalities than getstr() does.
    """
    while True:
        y, x = stdscr.getyx()
        ch = stdscr.getch()
        if ch == 10:  # Enter to complete
            break
        elif ch == curses.KEY_LEFT:
            # cursor remains still
            stdscr.move(y, indent if x <= indent else x - 1)
        elif ch == curses.KEY_RIGHT:
            stdscr.move(y, x + 1 if x < indent +
                        len(task['description']) else indent + len(task['description']))
        elif ch == curses.KEY_BACKSPACE or ch == 127:  # delete
            if x <= 7:
                stdscr.move(y, indent)  # cursor remains still
                continue
            # -1 because i am deleting the char before the cursor
            task['description'] = task['description'][:x -
                                                      indent - 1] + task['description'][x - indent:]
            tprint.print_task_mode(stdscr, task, y, mode)
            stdscr.move(y, x - 1)
        elif 32 <= ch < 127:  # printable char
            task['description'] = task['description'][:x - indent] + \
                chr(ch) + task['description'][x - indent:]
            tprint.print_task_mode(stdscr, task, y, mode)
            stdscr.move(y, x + 1)
        elif ch == 27 and mode == tprint.add_mode:  # todo: too slow
            return ""
    return task['description']


def set_color_selected(color: str):
    # invalid color
    if color not in color_set:
        return
    try:
        with open(tsk.settings_path, "r+") as settings_file:
            settings = json.load(settings_file)
            settings['task_highlighting_color'] = color
            settings_file.seek(0)  # move pointer back to beginning
            json.dump(settings, settings_file, indent=4)
            settings_file.truncate()
    except FileNotFoundError:
        setup_default_settings()


def get_color_selected():
    try:
        with open(tsk.settings_path, "r") as settings_file:
            settings = json.load(settings_file)
            color = settings['task_highlighting_color']
            return color_set[color]
    except FileNotFoundError:
        setup_default_settings()
        return curses.COLOR_BLUE


def setup_default_settings():
    """
    setup default settings if no settings.json were found
    """
    default_settings = {
        "date_format": "Y-M-D",
        "scroll": True,
        "autosort_f": False,
        "autosort_d": False,
        "task_highlighting_color": "blue"
    }
    with open(tsk.settings_path, "w") as file:
        json.dump(default_settings, file, indent=4)
    return default_settings

def is_view_fully_packed(start, end, max_count):
    """indicates whether the current view is completely filled with tasks"""
    return end - start + 1 >= max_count
    